/*****************************************
* UW User ID:  csubrama
* Submitted for ECE 250
* Department of Electrical and Computer Engineering
* University of Waterloo
* Calender Term of Submission: Winter 2018
*
* By submitting this file, I affirm that
* I am the author of all modifications to
* the provided code.
*****************************************/

//HEADER FILES DEFINED BELOW

#ifndef RPOVDGBQN9TIEO3P
#define RPOVDGBQN9TIEO3P

#include "Exception.h"
#include "ece250.h"
#include <cassert>

template <typename Type>
class Search_tree {
public:
	class Iterator;

private:
	class Node {
	public:
		Type node_value;
		int tree_height;

		// The left and right sub-trees
		Node *left_tree;
		Node *right_tree;

		// Hint as to how you can create your iterator
		// Point to the previous and next nodes in linear order
		Node *previous_node;
		Node *next_node;

		// Member functions
		Node(Type const & = Type());

		void update_height();

		int height() const;
		bool is_leaf() const;
		Node *front();
		Node *back();
		Node *find(Type const &obj);

		void clear();
		bool insert(Type const &obj, Node *&to_this);
		bool erase(Type const &obj, Node *&to_this);
	};

	Node *root_node;
	int tree_size;

	// Hint as to how to start your linked list of the nodes in order 
	Node *front_sentinel;
	Node *back_sentinel;

public:
	class Iterator {
	private:
		Search_tree * containing_tree;
		Node *current_node;
		bool is_end;

		// The constructor is private so that only the search tree can create an iterator
		Iterator(Search_tree *tree, Node *starting_node);

	public:
		// DO NOT CHANGE THE SIGNATURES FOR ANY OF THESE
		Type operator*() const;
		Iterator &operator++();
		Iterator &operator--();
		bool operator==(Iterator const &rhs) const;
		bool operator!=(Iterator const &rhs) const;

		// Make the search tree a friend so that it can call the constructor
		friend class Search_tree;
	};

	//display(): A custom method made by me to display the AVL tree onto the console after every insert

	//void display(Node *, int);

	// DO NOT CHANGE THE SIGNATURES FOR ANY OF THESE
	Search_tree();
	~Search_tree();

	bool empty() const;
	int size() const;
	int height() const;

	Type front() const;
	Type back() const;

	Iterator begin();
	Iterator end();
	Iterator rbegin();
	Iterator rend();
	Iterator find(Type const &);

	void clear();
	bool insert(Type const &);
	bool erase(Type const &);

	// Friends

	template <typename T>
	friend std::ostream &operator<<(std::ostream &, Search_tree<T> const &);
};

//////////////////////////////////////////////////////////////////////
//                Search Tree Public Member Functions               //
//////////////////////////////////////////////////////////////////////

// The initialization of the front and back sentinels is a hint

//Search_tree constructor 

template <typename Type>
Search_tree<Type>::Search_tree() :
	root_node(nullptr),
	tree_size(0),
	front_sentinel(new Search_tree::Node(Type())),
	back_sentinel(new Search_tree::Node(Type())) {
	front_sentinel->next_node = back_sentinel;
	back_sentinel->previous_node = front_sentinel;
}

//Search_tree destructor

template <typename Type>
Search_tree<Type>::~Search_tree() {
	clear();  // might as well use it...
	delete front_sentinel;
	delete back_sentinel;
}

//empty() function: Returns true if root node is '0' (nullptr)

template <typename Type>
bool Search_tree<Type>::empty() const {
	return (root_node == nullptr);
}

//size() fucntion: Returns the size of the tree

template <typename Type>
int Search_tree<Type>::size() const {
	return tree_size;
}

//height() fucntion: A recursive function that returns height from the root node!

template <typename Type>
int Search_tree<Type>::height() const {
	return root_node->height();
}

/* front() fucntion: Recursive function that returns the value of the front node from root.
                    Returns an underflow exception if empty!
*/

template <typename Type>
Type Search_tree<Type>::front() const {
	if (empty()) {
		throw underflow();
	}

	return root_node->front()->node_value;
}


/* back() fucntion: Recursive function that returns the value of the back node from root.
					Returns an underflow exception if empty!
*/

template <typename Type>
Type Search_tree<Type>::back() const {
	if (empty()) {
		throw underflow();
	}

	return root_node->back()->node_value;
}


// display(): Custom method made by me to display tree contents after every insert!

/*template <typename Type>
void Search_tree<Type>::display(Node *current, int indent) {

	if (current != nullptr) {
		display(current->left_tree, indent + 4);
		if (indent > 0)
			std::cout << std::setw(indent) << " ";
		std::cout << current->node_value << std::endl;
		display(current->right_tree, indent + 4);
		}
}*/

template <typename Type>
typename Search_tree<Type>::Iterator Search_tree<Type>::begin() {
	return empty() ? Iterator(this, back_sentinel) : Iterator(this, root_node->front());
}

template <typename Type>
typename Search_tree<Type>::Iterator Search_tree<Type>::end() {
	return Iterator(this, back_sentinel);
}

template <typename Type>
typename Search_tree<Type>::Iterator Search_tree<Type>::rbegin() {
	return empty() ? Iterator(this, front_sentinel) : Iterator(this, root_node->back());
}

template <typename Type>
typename Search_tree<Type>::Iterator Search_tree<Type>::rend() {
	return Iterator(this, front_sentinel);
}

template <typename Type>
typename Search_tree<Type>::Iterator Search_tree<Type>::find(Type const &obj) {
	if (empty()) {
		return Iterator(this, back_sentinel);
	}

	typename Search_tree<Type>::Node *search_result = root_node->find(obj);

	if (search_result == nullptr) {
		return Iterator(this, back_sentinel);
	}
	else {
		return Iterator(this, search_result);
	}
}

template <typename Type>
void Search_tree<Type>::clear() {
	if (!empty()) {
		root_node->clear();
		root_node = nullptr;
		tree_size = 0;
	}

	// Reinitialize the sentinels
	front_sentinel->next_node = back_sentinel;
	back_sentinel->previous_node = front_sentinel;
}

//insert(): This function inserts a node into the tree (as a child) depending on the root node
template <typename Type>
bool Search_tree<Type>::insert(Type const &obj) {
	if (empty()) {

		root_node = new Search_tree::Node(obj);

		//Update next_node of front sentinel & previous_node of root 

		front_sentinel->next_node = root_node;
		root_node->previous_node = front_sentinel;

		//Update previous_node of back sentinel & next_node of root 

		back_sentinel->previous_node = root_node;
		root_node->next_node = back_sentinel;

		//tree_size is 1 since only root node is inserted!

		tree_size = 1;

		//display(root_node, tree_size);
		//std::cout << std::endl;


		return true;
	}
	else if (root_node->insert(obj, root_node)) {

		++tree_size;

		//display(root_node, tree_size);
		//std::cout << std::endl;

		return true;

	}
	else {

		return false;

	}

}

//erase(): This function removes (erases) the node having node_value equal to the passed argument
template <typename Type>
bool Search_tree<Type>::erase(Type const &obj) {
	if (!empty() && root_node->erase(obj, root_node)) {
		--tree_size;
		return true;
	}
	else {
		return false;
	}
}

//////////////////////////////////////////////////////////////////////
//                   Node Public Member Functions                   //
//////////////////////////////////////////////////////////////////////

//Node class constructor
template <typename Type>
Search_tree<Type>::Node::Node(Type const &obj) :
	node_value(obj),
	left_tree(nullptr),
	right_tree(nullptr),
	next_node(nullptr),
	previous_node(nullptr),
	tree_height(0) {
	// does nothing
}

//update_height(): Returns maximum of (left or right sub tree height) + 1 (root node)
template <typename Type>
void Search_tree<Type>::Node::update_height() {
	tree_height = std::max(left_tree->height(), right_tree->height()) + 1;
}

//height(): Returns tree_height variable. Returns -1 if this is '0' (nullptr)
template <typename Type>
int Search_tree<Type>::Node::height() const {
	return (this == nullptr) ? -1 : tree_height;
}

// Return true if the current node is a leaf node, false otherwise
template <typename Type>
bool Search_tree<Type>::Node::is_leaf() const {
	return ((left_tree == nullptr) && (right_tree == nullptr));
}

// Return a pointer to the front node
template <typename Type>
typename Search_tree<Type>::Node *Search_tree<Type>::Node::front() {
	return (left_tree == nullptr) ? this : left_tree->front();
}

// Return a pointer to the back node
template <typename Type>
typename Search_tree<Type>::Node *Search_tree<Type>::Node::back() {
	return (right_tree == nullptr) ? this : right_tree->back();
}

/*find(): recursive function that returns the node in the sub trees contaning a node_value equal to that 
         of the passed argument! 
*/
template <typename Type>
typename Search_tree<Type>::Node *Search_tree<Type>::Node::find(Type const &obj) {
	if (obj == node_value) {
		return this;
	}
	else if (obj < node_value) {
		return (left_tree == nullptr) ? nullptr : left_tree->find(obj);
	}
	else {
		return (right_tree == nullptr) ? nullptr : right_tree->find(obj);
	}
}

// Recursively clear the tree
template <typename Type>
void Search_tree<Type>::Node::clear() {
	if (left_tree != nullptr) {
		left_tree->clear();
	}

	if (right_tree != nullptr) {
		right_tree->clear();
	}

	delete this;
}

/*insert(): This method inserts the node into the tree depending on a set of conditions that determine it's final postion in the tree
            We call these rotations. So depending on it's value, a set of rotations will be performed to maintain the balnce of the tree!
*/
template <typename Type>
bool Search_tree<Type>::Node::insert(Type const &obj, Search_tree<Type>::Node *&to_this) {

	// If obj < node_value, all operations will happen in the LEFT sub tree!

	if (obj < node_value) {
		if (left_tree == nullptr) {

			//If this is the first element being inserted in the left sub-tree

			left_tree = new Search_tree<Type>::Node(obj);

			//Updating previous_node & next_node to point to new node, i.e insert node to the tree

			previous_node->next_node = left_tree;
			left_tree->previous_node = previous_node;

			//Updating the inserted node as previous_node and next_node to point to itself!

			previous_node = left_tree;
			left_tree->next_node = to_this;

			//Update height!

			update_height();

			//no balancing needed!

			return true;
		}
		else {
			if (left_tree->insert(obj, left_tree)) {

				/*This sections involves checking if the tree is balanced or not. That is if the difference between the heights 
				  of the left sub tree and right sub tree is equal to '2', then the tree is unbalanced. Else it is balanced!		
				*/

				if (right_tree == nullptr || (left_tree->height() - right_tree->height() == 2)) {

					/* In this case the tree is unbalanced. So we check if we are going to perform a LEFT-LEFT or
					   or LEFT-RIGHT rotation!
					*/

					if (left_tree->left_tree->height() > left_tree->right_tree->height()) {

						//A LEFT-LEFT rotation is being performed by rotating the two nodes around the unbalanced node 'this'!

						Search_tree<Type>::Node *left = left_tree;
						Search_tree<Type>::Node *left_right = left->right_tree;
						left->right_tree = this;
						to_this = left;
						left_tree = left_right;

						// So this insertion will fix the balance of any ancestors all the way upto the root!
					}
					else if (left_tree->left_tree->height() < left_tree->right_tree->height()) {

						//A LEFT-RIGHT rotation is being performed by involving a series of 2 rotations! 

						/*This is achieved by assigning unbalanced root node and 'this' as children of a newlt created root node
                          and by connecting the two subtress and original parent of 'this'! 
						*/
						Search_tree<Type>::Node *left = left_tree;
						Search_tree<Type>::Node *left_right = left->right_tree;
						Search_tree<Type>::Node *left_right_left = left_right->left_tree;
						Search_tree<Type>::Node *left_right_right = left_right->right_tree;

						//Here is where we connect the two subtress...

						left_right->left_tree = left;
						left_right->right_tree = this;
						to_this = left_right;
						left->right_tree = left_right_left;
						left_tree = left_right_right;

						// So this insertion will fix the balance of any ancestors all the way upto the root!

					}
					else {

						//do nothing :)
					}
				}
				
				//Update height of the tree after insertion!

				update_height();
				return true;
			}
			else {
				return false;
			}
		}
	}
	else if (obj > node_value) {

		// If obj > node_value, all operations will happen in the RIGHT sub tree!

		if (right_tree == nullptr) {

			//If this is the first element being inserted in the right sub-tree

			right_tree = new Search_tree<Type>::Node(obj);

			//Updating previous_node & next_node to point to new node, i.e insert node to the tree

			next_node->previous_node = right_tree;
			right_tree->next_node = next_node;

			//Updating the inserted node as next_node and previous_node to point to itself!

			next_node = right_tree;
			right_tree->previous_node = to_this;

			//Update height!

			update_height();

			//no balancing needed!

			return true;
		}
		else {
			if (right_tree->insert(obj, right_tree)) {

				/*This sections involves checking if the tree is balanced or not. That is if the difference between the heights
				of the right sub tree and left sub tree is equal to '2', then the tree is unbalanced. Else it is balanced!
				*/

				if (left_tree == nullptr || (right_tree->height() - left_tree->height() == 2)) {

					/* In this case the tree is unbalanced. So we check if we are going to perform a RIGHT-RIGHT or
					or RIGHT-LEFT rotation!
					*/

					if (right_tree->right_tree->height() > right_tree->left_tree->height()) {

						/*A RIGHT-RIGHT rotation is being performed by rotating the two nodes around the unbalanced node 'this'!
						  Note that this is symmetrical to a LEFT-LEFT rotation!
						*/

						Search_tree<Type>::Node *right = right_tree;
						Search_tree<Type>::Node *right_left = right->left_tree;
						right->left_tree = this;
						to_this = right;
						right_tree = right_left;

						// So this insertion will fix the balance of any ancestors all the way upto the root!

					}
					else if (right_tree->right_tree->height() < right_tree->left_tree->height()) {

						//A RIGHT-LEFT rotation is being performed by involving a series of 2 rotations! 

						/*This is achieved by assigning unbalanced root node and 'this' as children of a newlt created root node
						and by connecting the two subtress and original parent of 'this'!
						*/

						//Note that this is symmetrical to a LEFT-RIGHT rotation!

						Search_tree<Type>::Node *right = right_tree;
						Search_tree<Type>::Node *right_left = right->left_tree;
						Search_tree<Type>::Node *right_left_right = right_left->right_tree;
						Search_tree<Type>::Node *right_left_left = right_left->left_tree;

						//So here is where we connect the two subtrees!

						right_left->right_tree = right;
						right_left->left_tree = this;
						to_this = right_tree;
						right->left_tree = right_left_right;
						right_tree = right_left_left;

						// So this insertion will fix the balance of any ancestors all the way upto the root!

					}
					else {

						//do nothing :)
					}
				}

				//Update height of the tree after insertion!
				
				update_height();
				return true;
			}
			else {
				return false;
			}
		}
	}
	else {
		return false;
	}
}

template <typename Type>
bool Search_tree<Type>::Node::erase(Type const &obj, Search_tree<Type>::Node *&to_this) {

	//Note that erase can have multiple im-balances. So after removal we must ensure all ancestors are balanced upto the root!

	// If obj < node_value, all erase operations & rotations will happen in the RIGHT sub tree!

	if (obj < node_value) {
		if (left_tree == nullptr) {
			return false;
		}
		else {
			if (left_tree->erase(obj, left_tree)) {

				/*This sections involves checking if the tree is balanced or not. That is if the difference between the heights
				of the right sub tree and left sub tree is equal to '2', then the tree is unbalanced. Else it is balanced!
				*/

				if (right_tree->height() - left_tree->height() == 2) {

					/* In this case the tree is unbalanced. So we check if we are going to perform a RIGHT-RIGHT or
					or RIGHT-LEFT rotation!
					*/

					if (right_tree->right_tree->height() > right_tree->left_tree->height()) {

						/*A RIGHT-RIGHT rotation is being performed by rotating the two nodes around the unbalanced node 'this'!
						Note that this is symmetrical to a LEFT-LEFT rotation!
						*/

						Search_tree<Type>::Node *right = right_tree;
						Search_tree<Type>::Node *right_left = right->left_tree;
						right->left_tree = this;
						to_this = right;
						right_tree = right_left;

						// So this insertion will fix the balance of any ancestors all the way upto the root!

					}
					else if (right_tree->right_tree->height() < right_tree->left_tree->height()) {

						//A RIGHT-LEFT rotation is being performed by involving a series of 2 rotations! 

						/*This is achieved by assigning unbalanced root node and 'this' as children of a newlt created root node
						and by connecting the two subtress and original parent of 'this'!
						*/

						//Note that this is symmetrical to a LEFT-RIGHT rotation!

						Search_tree<Type>::Node *right = right_tree;
						Search_tree<Type>::Node *right_left = right->left_tree;
						Search_tree<Type>::Node *right_left_right = right_left->right_tree;
						Search_tree<Type>::Node *right_left_left = right_left->left_tree;

						//So here is where we connect the two subtrees!

						right_left->right_tree = right;
						right_left->left_tree = this;
						to_this = right_tree;
						right->left_tree = right_left_right;
						right_tree = right_left_left;

						// So this insertion will fix the balance of any ancestors all the way upto the root!

					}
					else {

						//do nothing :)
					}
				}
				update_height();
				return true;
			}

			return false;
		}
	}
	else if (obj > node_value) {

		// If obj > node_value, all erase operations & rotations will happen in the LEFT sub tree!

		if (right_tree == nullptr) {
			return false;
		}
		else {
			if (right_tree->erase(obj, right_tree)) {

				/*This sections involves checking if the tree is balanced or not. That is if the difference between the heights
				of the right sub tree and left sub tree is equal to '2', then the tree is unbalanced. Else it is balanced!
				*/

				if (left_tree->height() - right_tree->height() == 2) {

					/* In this case the tree is unbalanced. So we check if we are going to perform a LEFT-LEFT or
					or LEFT-RIGHT rotation!
					*/

					if (left_tree->left_tree->height() > left_tree->right_tree->height()) {

						//A LEFT-LEFT rotation is being performed by rotating the two nodes around the unbalanced node 'this'!

						Search_tree<Type>::Node *left = left_tree;
						Search_tree<Type>::Node *left_right = left->right_tree;
						left->right_tree = this;
						to_this = left;
						left_tree = left_right;

						// So this insertion will fix the balance of any ancestors all the way upto the root!
					}
					else if (left_tree->left_tree->height() < left_tree->right_tree->height()) {

						//A LEFT-RIGHT rotation is being performed by involving a series of 2 rotations! 

						/*This is achieved by assigning unbalanced root node and 'this' as children of a newlt created root node
						and by connecting the two subtress and original parent of 'this'!
						*/
						Search_tree<Type>::Node *left = left_tree;
						Search_tree<Type>::Node *left_right = left->right_tree;
						Search_tree<Type>::Node *left_right_left = left_right->left_tree;
						Search_tree<Type>::Node *left_right_right = left_right->right_tree;

						//Here is where we connect the two subtress...

						left_right->left_tree = left;
						left_right->right_tree = this;
						to_this = left_right;
						left->right_tree = left_right_left;
						left_tree = left_right_right;

						// So this insertion will fix the balance of any ancestors all the way upto the root!

					}
					else {

						//do nothing :)
					}
				}
				update_height();
				return true;
			}

			return false;
		}
	}
	else {
		assert(obj == node_value);

		//if it's a leaf node, after removal update previous and next node values!

		if (is_leaf()) {
			next_node->previous_node = previous_node;
			previous_node->next_node = next_node;
			to_this = nullptr;
			delete this;
		}
		else if (left_tree == nullptr) {

			//if left_tree is '0', after removal update previous and next node values!

			next_node->previous_node = previous_node;
			previous_node->next_node = next_node;
			to_this = right_tree;
			delete this;
		}
		else if (right_tree == nullptr) {

			//if right_tree is '0', after removal update previous and next node values!

			next_node->previous_node = previous_node;
			previous_node->next_node = next_node;
			to_this = left_tree;
			delete this;
		}
		else {

			//After removal update previous and next node values!

			node_value = right_tree->front()->node_value;
			right_tree->erase(node_value, right_tree);
			update_height();
		}

		return true;
	}
}

//////////////////////////////////////////////////////////////////////
//                   Iterator Private Constructor                   //
//////////////////////////////////////////////////////////////////////

template <typename Type>
Search_tree<Type>::Iterator::Iterator(Search_tree<Type> *tree, typename Search_tree<Type>::Node *starting_node) :
	containing_tree(tree),
	current_node(starting_node) {
	// This is done for you...
	// Does nothing...
}

//////////////////////////////////////////////////////////////////////
//                 Iterator Public Member Functions                 //
//////////////////////////////////////////////////////////////////////

template <typename Type>
Type Search_tree<Type>::Iterator::operator*() const {
	// This is done for you...
	return current_node->node_value;
}

template <typename Type>
typename Search_tree<Type>::Iterator &Search_tree<Type>::Iterator::operator++() {
	// Update the current node to the node containing the next higher value
	// If we are already at end do nothing

	// Your implementation here, do not change the return value

	if (current_node != containing_tree->back_sentinel) {
		current_node = current_node->next_node;
	}

	return *this;
}

template <typename Type>
typename Search_tree<Type>::Iterator &Search_tree<Type>::Iterator::operator--() {
	// Update the current node to the node containing the next smaller value
	// If we are already at either rend, do nothing

	// Your implementation here, do not change the return value

	if (current_node != containing_tree->front_sentinel) {
		current_node = current_node->previous_node;
	}

	return *this;
}

template <typename Type>
bool Search_tree<Type>::Iterator::operator==(typename Search_tree<Type>::Iterator const &rhs) const {
	// This is done for you...
	return (current_node == rhs.current_node);
}

template <typename Type>
bool Search_tree<Type>::Iterator::operator!=(typename Search_tree<Type>::Iterator const &rhs) const {
	// This is done for you...
	return (current_node != rhs.current_node);
}

//////////////////////////////////////////////////////////////////////
//                            Friends                               //
//////////////////////////////////////////////////////////////////////

// You can modify this function however you want:  it will not be tested

template <typename T>
std::ostream &operator<<(std::ostream &out, Search_tree<T> const &list) {
	out << "not yet implemented";

	return out;
}

#endif
