/*****************************************
 * UW User ID:  csubrama
 * Submitted for ECE 250
 * Department of Electrical and Computer Engineering
 * University of Waterloo
 * Calender Term of Submission: Winter 2018
 *
 * By submitting this file, I affirm that
 * I am the author of all modifications to
 * the provided code.
 *****************************************/

#ifndef DOUBLE_SENTINEL_LIST_H
#define DOUBLE_SENTINEL_LIST_H

#include <iostream>
#include "Exception.h"

 /*
 * Double_sentinel_list
 *
 * Modified by: Chandrakiran Subramaniam
 * Date of completion: 2018-01-14, 00:51:46 EST
 *
 * Below is a summary of all functions associated with this class.
 * You will notice that it has been documented in the format of an
 * UML class diagram. Comments have been included before every func.
 * implementation for the ease understanding :)
 *
	 - list_head:Double_node
	 - list_tail:Double_node
	 - list_size:Integer
	 + create():Double_sentinel_list
	 + create( in dsl:Double_sentinel_list ):Double_sentinel_list
	 + size():Integer
	 + empty():Boolean
	 + front():Type
	 + back():Type
	 + begin():Double_node
	 + end():Double_node
	 + rbegin():Double_node
	 + rend():Double_node
	 + find( in obj:Type ):Double_node
	 + count( in obj:Type ):Integer
	 + swap( inout list:Double_sentinel_list )
	 + =( in rhs:Double_sentinel_list ):Double_sentinel_list
	 + push_front( in obj:Type )
	 + push_back( in obj:Type )
	 + pop_front()
	 + pop_back():Type
	 + erase( in obj:Type ):Integer
	 + destroy()
 */

template <typename Type>
class Double_sentinel_list {
	public:
		class Double_node {
			public:
				Double_node( Type const & = Type(), Double_node * = nullptr, Double_node * = nullptr );

				Type value() const;
				Double_node *previous() const;
				Double_node *next() const;

				Type         node_value;
				Double_node *previous_node;
				Double_node *next_node;
		};

		Double_sentinel_list();
		Double_sentinel_list( Double_sentinel_list const & );
		Double_sentinel_list( Double_sentinel_list && );
		~Double_sentinel_list();

		// Accessors

		int size() const;
		bool empty() const;

		Type front() const;
		Type back() const;

		Double_node *begin() const;
		Double_node *end() const;
		Double_node *rbegin() const;
		Double_node *rend() const;

		Double_node *find( Type const & ) const;
		int count( Type const & ) const;

		// Mutators

		void swap( Double_sentinel_list & );
		Double_sentinel_list &operator=( Double_sentinel_list );
		Double_sentinel_list &operator=( Double_sentinel_list && );

		void push_front( Type const & );
		void push_back( Type const & );

		void pop_front();
		void pop_back();

		int erase( Type const & );

	private:
		Double_node *list_head;
		Double_node *list_tail;
		int list_size;

		// List any additional private member functions you author here
	// Friends

	template <typename T>
	friend std::ostream &operator<<( std::ostream &, Double_sentinel_list<T> const & );
};

/////////////////////////////////////////////////////////////////////////
//                      Public member functions                        //
/////////////////////////////////////////////////////////////////////////


/**************************************************************************
 * * *                        SUB SECTION - 1                         * * *
 * * *           CONSTRUCTORS & DESTRUCTORS IMPLEMENTATION            * * *
***************************************************************************/

/*
* Function: Double_sentinel_list()
* Parameters: (Type(), nullptr, nullptr)
* Return type: NO RETURN TYPE
*
* Description: Initializes the list with a head and tail pointer 
*              pointing to two, dynamically allocated, sentinel 
*              nodes.
*/

template <typename Type>
Double_sentinel_list<Type>::Double_sentinel_list():
// Updated the initialization list here
list_head(nullptr),
list_tail(nullptr),
list_size( 0 )
{
	//Creating two head and tail sentinel nodes and pointed by the head and tail pointers

	Double_node *sentinel_head = new Double_node();
	Double_node *sentinel_tail = new Double_node();
	list_head = sentinel_head;
	list_tail = sentinel_tail;

	//Implementing an empty list

	list_head->next_node = list_tail;
	list_tail->previous_node = list_head;

}

/*
* Function: Double_sentinel_list(Double_sentinel_list const &list)
* Parameters: (Double_sentinel_list const &list)
* Return type: NO RETURN TYPE
*
* Description: The copy constructor must create a new doubly linked
*              list with a copy of all of the nodes within the 
*              linked list pass as the argument list with the values
*              stored in the same order. 
*/

template <typename Type>
Double_sentinel_list<Type>::Double_sentinel_list( Double_sentinel_list<Type> const &list ):
// Updated the initialization list here
list_head(nullptr),
list_tail(nullptr),
list_size( 0 )
{

	list_head = new Double_node();
	list_tail = new Double_node();

	list_head->next_node = list_tail;
	list_tail->previous_node = list_head;

	//Iterating through the passed list and copying all elements

	for (
		Double_node *ptr = list.begin();
		ptr != list.end();
		ptr = ptr->next()
		) {	
		push_back(ptr->node_value);
	}

}

/*
* Function: Double_sentinel_list( Double_sentinel_list<Type> &&list )
* Parameters: (Double_sentinel_list<Type> &&list)
* Return type: NO RETURN TYPE
*
* Description: The move constructor must create a new doubly linked
*              list with all the nodes found within the linked list
*              passed as an argument list with the values stored in
*              the same order.
*/

template <typename Type>
Double_sentinel_list<Type>::Double_sentinel_list( Double_sentinel_list<Type> &&list ):
// Updated the initialization list here
list_head( nullptr ),
list_tail( nullptr ),
list_size( 0 )
{

	list_head = new Double_node();
	list_tail = new Double_node();

	list_head->next_node = list_tail;
	list_tail->previous_node = list_head;

	//Calling swap function to swap head & tail of the list

	this->swap(list);
}

/*
* Function: ~Double_sentinel_list()
* Parameters: NONE
* Return type: NO RETURN TYPE
*
* Description: The destructor deletes each of the nodes in the list
*              including the sentinels.
*/

template <typename Type>
Double_sentinel_list<Type>::~Double_sentinel_list() {

	while (!empty()) {
		pop_front();
	}

	delete list_head;
	delete list_tail;
}

/**************************************************************************
* * *                        SUB SECTION - 2                         * * *
* * *                   ACCESSORS IMPLEMENTATION                     * * *
***************************************************************************/

/*
* Function: int Double_sentinel_list<Type>::size() const
* Parameters: NONE
* Return type: INT
*
* Description: Returns the number of items in the list.
*/

template <typename Type>
int Double_sentinel_list<Type>::size() const {

	return list_size;
}

/*
* Function: bool Double_sentinel_list<Type>::empty() const
* Parameters: NONE
* Return type: BOOL
*
* Description: Returns true if the list is empty, false otherwise.
*/

template <typename Type>
bool Double_sentinel_list<Type>::empty() const {
	
	if (list_head->next() == list_tail) {
		return true;
	}
	else
		return false;
}

/*
* Function: Type Double_sentinel_list<Type>::front() const
* Parameters: NONE
* Return type: TYPE()
*
* Description: Retrieves the object stored in the node pointed to
*              by the next pointer of the head sentinel.
*/

template <typename Type>
Type Double_sentinel_list<Type>::front() const {

	if (empty()) {
		throw underflow();
	}
	else
		return list_head->next()->value();
}

/*
* Function: Type Double_sentinel_list<Type>::back() const
* Parameters: NONE
* Return type: TYPE()
*
* Description: Retrieves the object stored in the node pointed to
*              by the next pointer of the tail sentinel.
*/

template <typename Type>
Type Double_sentinel_list<Type>::back() const {

	if (empty()) {
		throw underflow();
	}
	else
		return list_tail->previous()->value();
}

/*
* Function: typename Double_sentinel_list<Type>::Double_node *Double_sentinel_list<Type>::begin() const
* Parameters: NONE
* Return type: typename
*
* Description: Returns the address stored by the next pointer of 
*              the head sentinel node
*/

template <typename Type>
typename Double_sentinel_list<Type>::Double_node *Double_sentinel_list<Type>::begin() const {

	return list_head->next_node;
}

/*
* Function: typename Double_sentinel_list<Type>::Double_node *Double_sentinel_list<Type>::end() const
* Parameters: NONE
* Return type: typename
*
* Description: Returns the address of the tail sentinel node.
*/

template <typename Type>
typename Double_sentinel_list<Type>::Double_node *Double_sentinel_list<Type>::end() const {

	return list_tail;
}

/*
* Function: typename Double_sentinel_list<Type>::Double_node *Double_sentinel_list<Type>::rbegin() const
* Parameters: NONE
* Return type: typename
*
* Description: Returns the address stored by the previous pointer of the tail 
*              sentinel node. 
*/

template <typename Type>
typename Double_sentinel_list<Type>::Double_node *Double_sentinel_list<Type>::rbegin() const {

	return list_tail->previous_node;
}

/*
* Function: typename Double_sentinel_list<Type>::Double_node *Double_sentinel_list<Type>::rend() const
* Parameters: NONE
* Return type: typename
*
* Description: Returns the address of the head sentinel node.
*/

template <typename Type>
typename Double_sentinel_list<Type>::Double_node *Double_sentinel_list<Type>::rend() const {

	return list_head;
}

/*
* Function: typename Double_sentinel_list<Type>::Double_node *Double_sentinel_list<Type>::find( Type const &obj ) const
* Parameters: ( Type const &obj )
* Return type: typename
*
* Description: Returns the address of the first node in the linked list storing
*              a value equal to the argument; if none is found, return end()
*/

template <typename Type>
typename Double_sentinel_list<Type>::Double_node *Double_sentinel_list<Type>::find( Type const &obj ) const {
	
	//Traversing list and returning address of the node containing value equal to that of the argument passed

	for (auto *ptr = begin(); ptr != end(); ptr = ptr->next()) {
		if (ptr->value() == obj) {
			return ptr;
		}
	}
	return end();
}

/*
* Function: int Double_sentinel_list<Type>::count( Type const &obj ) const
* Parameters: ( Type const &obj )
* Return type: INT
*
* Description: Returns the number of nodes in the linked list storing
*              a value equal to the argument. 
*/

template <typename Type>
int Double_sentinel_list<Type>::count( Type const &obj ) const {

	int num=0;
	Double_node *ptr = list_head->next();

	//Traversing the list and counting the number of nodes storing the value equal to that of the passed argument value

	while (ptr != list_tail) {
		Double_node *next = ptr->next();
		if (ptr->node_value == obj) {
			num++;
		}
		ptr = next;
	}
	return num;
}

/*********************************************************************
* *                       SUB SECTION - 3                         * *
* *                  MUTATORS IMPLEMENTATION                      * *
*********************************************************************/

template <typename Type>
void Double_sentinel_list<Type>::swap( Double_sentinel_list<Type> &list ) {
	// This is done for you
	std::swap( list_head, list.list_head );

	std::swap( list_tail, list.list_tail );
	std::swap( list_size, list.list_size );
}

template <typename Type>
Double_sentinel_list<Type> &Double_sentinel_list<Type>::operator=( Double_sentinel_list<Type> rhs ) {
	// This is done for you
	swap( rhs );

	return *this;
}

template <typename Type>
Double_sentinel_list<Type> &Double_sentinel_list<Type>::operator=( Double_sentinel_list<Type> &&rhs ) {
	// This is done for you
	swap( rhs );

	return *this;
}

/*
* Function: void Double_sentinel_list<Type>::push_front( Type const &obj )
* Parameters: ( Type const &obj )
* Return type: VOID
*
* Description: Pushes a new node in the front of the linked list storing
*              a value equal to the argument passed.
*/

template <typename Type>
void Double_sentinel_list<Type>::push_front( Type const &obj ) {
	
	//Making a temporary node intended to be pushed to the list

	Double_node *tmp = new Double_node(obj, nullptr, nullptr);

	//Make a connection to the list by assigning the value, previous and next pointers
	
	tmp->next_node = list_head->next();
	tmp->previous_node = list_head;
	tmp->node_value = obj;

	//Point elements before and after the node to the node itself

	tmp->next()->previous_node = tmp;
	tmp->previous()->next_node = tmp;

	list_size++;

}

/*
* Function: void Double_sentinel_list<Type>::push_back( Type const &obj )
* Parameters: ( Type const &obj )
* Return type: VOID
*
* Description: Pushes a new node in the back of the linked list storing
*              a value equal to the argument passed.
*/

template <typename Type>
void Double_sentinel_list<Type>::push_back( Type const &obj ) {
	
	//Making a temporary node intended to be pushed to the list

	Double_node *tmp = new Double_node(obj, nullptr, nullptr);

	//Make a connection to the list by assigning the value, previous and next pointers

	tmp->previous_node = list_tail->previous();
	tmp->next_node = list_tail;
	tmp->node_value = obj;

	//Point elements before and after the node to the node itself

	tmp->next()->previous_node = tmp;
	tmp->previous()->next_node = tmp;

	list_size++;
}

/*
* Function: Double_sentinel_list<Type>::pop_front()
* Parameters: NONE
* Return type: VOID
*
* Description: Pops a node in the front of the linked list.
*/

template <typename Type>
void Double_sentinel_list<Type>::pop_front() {
	
	if (empty()) {
		throw underflow();
	}

	//Storing element to be popped in a temp node

	Double_node *tmp = list_head->next();

	//Operation making the head sentinel to point to next element

	tmp->previous()->next_node = tmp->next();
	tmp->next()->previous_node = tmp->previous();

	delete tmp;

	list_size--;

}

/*
* Function: Double_sentinel_list<Type>::pop_back()
* Parameters: NONE
* Return type: VOID
*
* Description: Pops a node at the back of the linked list.
*/

template <typename Type>
void Double_sentinel_list<Type>::pop_back() {
	// Enter your implementation here
	if (empty()) {
		throw underflow();
	}

	//Storing element to be popped in a temp node

	Double_node *tmp = list_tail->previous();

	//Operation making the tail sentinel to point to previous element

	tmp->previous()->next_node = tmp->next();
	tmp->next()->previous_node = tmp->previous();

	delete tmp;

	list_size--;

}

/*
* Function: int Double_sentinel_list<Type>::erase( Type const &obj )
* Parameters: ( Type const &obj )
* Return type: INT
*
* Description: Delete all the nodes in the linked list that have a value
*              equal to the argument value.
*/

template <typename Type>
int Double_sentinel_list<Type>::erase( Type const &obj ) {

	//Storing first element in a temp node

	int num = 0;
	Double_node * tmp = list_head->next();

	while (tmp != list_tail) {
		Double_node *next = tmp->next();

		//IF statement that compares and deletes node equal to value passed in the argument

		if (tmp->node_value == obj) {
			list_size--; num++;
			tmp->previous()->next_node = tmp->next();
			tmp->next()->previous_node = tmp->previous();
			delete tmp;
		}
		tmp = next;
	}
	return num;
}

template <typename Type>
Double_sentinel_list<Type>::Double_node::Double_node(
	Type const &nv,
	typename Double_sentinel_list<Type>::Double_node *pn,
	typename Double_sentinel_list<Type>::Double_node *nn ):
// Updated the initialization list here
node_value( Type() ), // This assigns 'node_value' the default value of Type
previous_node( nullptr ),
next_node( nullptr )
{
	// Enter your implementation here

	node_value = nv;
	previous_node = pn;
	next_node = nn;
}

template <typename Type>
Type Double_sentinel_list<Type>::Double_node::value() const {
	
	//returns the value stored in the node

	return node_value; 
}

template <typename Type>
typename Double_sentinel_list<Type>::Double_node *Double_sentinel_list<Type>::Double_node::previous() const {

	//return the previous nodes' address
	
	return previous_node;
}

template <typename Type>
typename Double_sentinel_list<Type>::Double_node *Double_sentinel_list<Type>::Double_node::next() const {
	
	//Returns the next nodes' address

	return next_node;
}

/////////////////////////////////////////////////////////////////////////
//                      Private member functions                       //
/////////////////////////////////////////////////////////////////////////

// If you author any additional private member functions, include them here

/////////////////////////////////////////////////////////////////////////
//                               Friends                               //
/////////////////////////////////////////////////////////////////////////

// You can modify this function however you want:  it will not be tested

template <typename T>
std::ostream &operator<<( std::ostream &out, Double_sentinel_list<T> const &list ) {
	out << "head";

	for ( typename Double_sentinel_list<T>::Double_node *ptr = list.rend(); ptr != nullptr; ptr = ptr->next() ) {
		if ( ptr == list.rend() || ptr == list.end() ) {
			out << "->S";
		} else {
			out << "->" << ptr->value();
		}
	}

	out << "->0" << std::endl << "tail";

	for ( typename Double_sentinel_list<T>::Double_node *ptr = list.end(); ptr != nullptr; ptr = ptr->previous() ) {
		if ( ptr == list.rend() || ptr == list.end() ) {
			out << "->S";
		} else {
			out << "->" << ptr->value();
		}
	}

	out << "->0";

	return out;
}

#endif
