/*****************************************
* UW User ID:  csubrama
* Submitted for ECE 250
* Department of Electrical and Computer Engineering
* University of Waterloo
* Calender Term of Submission: Winter 2018
*
* By submitting this file, I affirm that
* I am the author of all modifications to
* the provided code.
*****************************************/

#ifndef DYNAMIC_DEQUE_H
#define DYNAMIC_DEQUE_H

#include "Exception.h"
#include<algorithm>

/*
* Resizable_deque
*
* Modified by: Chandrakiran Subramaniam
* Date of completion: 2018-01-29, 19:50:46 EST
*
* In this class, we implement various functions that will aid us
* in making a resizable deque. 
* 
* In this project, I have implemented a few custom private/ public
* functions to help in accessing the deque and resizing it as and
* when needed. Please find below a summary of the custom functions
* made by me...
*
* Custom functions made by Chandrakiran:
* 
* public:
*			int front_position();
*			int back_position();
*			Type retreive(int) const;
*
* private:
*			bool full() const;
*			void double_size();
*			void half_size();
*/

template <typename Type>
class Resizable_deque {
	public:
		Resizable_deque(int = 16);
		Resizable_deque( Resizable_deque const & );
		Resizable_deque( Resizable_deque && );
		~Resizable_deque();

		//Accessors of the deque

		Type front() const;
		Type back() const;
		int size() const;
		bool empty() const;
		int capacity() const;

		//Mutators of the deque

		void swap( Resizable_deque & );
		Resizable_deque &operator=( Resizable_deque const& );
		Resizable_deque &operator=( Resizable_deque && );
		void push_front( Type const & );
		void push_back( Type const & );
		void pop_front();
		void pop_back();
		void clear();

		//Custom public helper functions

		int front_position();
		int back_position();
		Type retreive(int) const;

	private:

		int deque_size;
		int ifront;
		int iback;
		int initial_array_capacity;
		int array_capacity;
		Type *array;

		//Custom private helper functions

		bool full() const;
		void double_size();
		void half_size();

	//Friends

	template <typename T>
	friend std::ostream &operator<<( std::ostream &, Resizable_deque<T> const & );
};

/**************************************************************************
* * *                        SUB SECTION - 1                         * * *
* * *           CONSTRUCTORS & DESTRUCTORS IMPLEMENTATION            * * *
***************************************************************************/

// Constructor
//Note the use of conditional operator to satisfy the logic of default size 16!
template <typename Type>
Resizable_deque<Type>::Resizable_deque( int n ):
	deque_size(0),
	ifront((n >= 16) ? n : 16),
	iback(-1),
	initial_array_capacity((n>=16) ? n:16),
	array_capacity(std::max(initial_array_capacity,n)),
	array(new Type [array_capacity]){

	//Initializing array to zero

	for (int i = 0; i < array_capacity; i++) {
		array[i] = 0;
	}

}

// Copy Constructor
template <typename Type>
Resizable_deque<Type>::Resizable_deque( Resizable_deque const &deque ):
	deque_size(0),
	ifront(deque.array_capacity),
	iback(-1),
	initial_array_capacity(deque.initial_array_capacity),
	array_capacity(std::max(initial_array_capacity, deque.array_capacity)),
	array(new Type[array_capacity]) {

	int ctr = 0;

	//Initializing array to zero

	for (int i = 0; i < array_capacity; i++) {
		array[i] = 0;
	}

	//Iterating through passed deque and copying all elements to array

	for (int i = 0; i < array_capacity; i++) {
		array[i] = deque.retreive(i);
		if (array[i] != 0) {
			ctr++;
		}
	}

	//Assigning front and back to the copied array

	ifront = deque.ifront;
	iback = deque.iback;

	deque_size = ctr;

}

// Move Constructor
template <typename Type>
Resizable_deque<Type>::Resizable_deque( Resizable_deque &&deque ):
	deque_size(0),
	ifront(deque.array_capacity),
	iback(-1),
	initial_array_capacity(deque.initial_array_capacity),
	array_capacity(std::max(initial_array_capacity, deque.array_capacity)),
	array(new Type[array_capacity]) {

	//Calling swap function to swap all member variables of the deuqe with array

	this->swap(deque);
}

// Destructor
template <typename Type>
Resizable_deque<Type>::~Resizable_deque() {
	
	delete[] array;
}

/**************************************************************************
* * *                        SUB SECTION - 2                         * * *
* * *                   ACCESSORS IMPLEMENTATION                     * * *
***************************************************************************/

/*
* Description: Returns the number of elements currently stored in the deque.
*/
template <typename Type>
int Resizable_deque<Type>::size() const {
	
	return deque_size;
}

/*
* Description: Returns the current capacity of the array.
*/
template <typename Type>
int Resizable_deque<Type>::capacity() const {
	
	return array_capacity;
}

/*
* Description: Returns true if the deque is empty, false otherwise.
*/
template <typename Type>
bool Resizable_deque<Type>::empty() const {
	
	return (deque_size == 0);
}

/*
* Description: Returns element in front of the deque, returns underflow if deque is empty.
*/
template <typename  Type>
Type Resizable_deque<Type>::front() const {
	
	if (empty()) {
		throw underflow();
	}

	return array[ifront];
}

/*
* Description: Returns element at back of the deque, returns underflow if deque is empty.
*/
template <typename  Type>
Type Resizable_deque<Type>::back() const {
	
	if (empty()) {
		throw underflow();
	}

	return array[iback];
}

/*********************************************************************
* *                       SUB SECTION - 3                         * *
* *                  MUTATORS IMPLEMENTATION                      * *
*********************************************************************/

/*
* Description: The swap function swaps all the member variables of this deque with those of 
*			   the argument.
*/
template <typename Type>
void Resizable_deque<Type>::swap( Resizable_deque<Type> &deque ) {
	
	//Swapping member variables

	std::swap(ifront, deque.ifront);
	std::swap(iback, deque.iback);
	std::swap(deque_size, deque.deque_size);
	std::swap(initial_array_capacity, deque.initial_array_capacity);
	std::swap(array_capacity, deque.array_capacity);

	//Swapping instance of the arrays

	std::swap(array, deque.array);

}

/*
* Description: The assignment operator makes a copy of the argument and then swaps the member variables
*              of this node with those of the copy.
*/
template <typename Type>
Resizable_deque<Type> &Resizable_deque<Type>::operator=( Resizable_deque<Type> const &rhs ) {
	
	Resizable_deque<Type> copy( rhs );
	swap( copy );

	return *this;
}

/*
* Description: The move operator moves the array in the argument (the right-hand side of the assignment) linked
*              list to this linked list, changing the argument linked list into an empty list.
*/
template <typename Type>
Resizable_deque<Type> &Resizable_deque<Type>::operator=( Resizable_deque<Type> &&rhs ) {
	
	swap( rhs );

	return *this;
}

/*
* Description: Insert the new element at the front of the deque. If before the element is placed into the deque, 
*              the array is filled, the capacity of the array is doubled.
*/
template <typename Type>
void Resizable_deque<Type>::push_front( Type const &obj ) {

	//Call double_size() if array is full!

	if (full()){
		double_size();
	}

	//In my implementation, we are pushing front from the back of the array, hence ifront--

	ifront--;

	//If ifront indexes a negative number, we loop back!

	if (ifront < 0) {
		ifront = (array_capacity - 1);
	}

	array[ifront] = obj;
	deque_size++;

	//If we just push_front the first element of the array, iback and ifront must point to it...

	if (deque_size==1) {
		iback = ifront;
	}
	
}

/*
* Description: Insert the new element at the back of the deque. If before the element is placed into the deque,
*              the array is filled, the capacity of the array is doubled.
*/
template <typename Type>
void Resizable_deque<Type>::push_back( Type const &obj ) {

	//Call double_size() if array is full!
	
	if (full()) {
		double_size();
	}

	//In my implementation, we are pushing back from the front of the array, hence iback++

	iback++;

	//If iback indexes a value out of array bounds, we loop front!

	if (iback > (array_capacity - 1)) {
		iback = 0;
	}

	array[iback] = obj;
	deque_size++;

	//If we just push_back the first element of the array, iback and ifront must point to it...

	if (deque_size == 1) {
		ifront = iback;
	}
}

/*
* Description: Removes the element at the front of the deque. If, after the element is removed, 
*              the array is 1/4 full and the array capacity is greater than the initial capacity, 
*              the capacity of the array is halved. This may throw a underflow exception. 
*/
template <typename Type>
void Resizable_deque<Type>::pop_front() {
	
	if (empty()) {
		throw underflow();
	}

	//Popping element by re-assigning it as zero...

	array[ifront] = 0;
	ifront++;
	deque_size--;

	//If after popping, ifront is out of array bounds, loop back!

	if (ifront > (array_capacity - 1)) {
		ifront = 0;
	}

	//Condition to implement half_size(), that results in an array half the size of the current one...

	if (deque_size == (array_capacity / 4) && (array_capacity>initial_array_capacity)) {
		half_size();
	}

}

/*
* Description: Removes the element at the back of the deque. If, after the element is removed,
*              the array is 1/4 full and the array capacity is greater than the initial capacity,
*              the capacity of the array is halved. This may throw a underflow exception.
*/
template <typename Type>
void Resizable_deque<Type>::pop_back() {
	
	if (empty()) {
		throw underflow();
	}

	//If before popping, iback is out of array bounds, loop front!

	if (iback < 0) {
		iback = array_capacity - 1;
	}

	//Popping element by re-assigning it as zero...

	array[iback] = 0;
	iback--;
	deque_size--;

	//Condition to implement half_size(), that results in an array half the size of the current one...

	if (deque_size == (array_capacity / 4) && (array_capacity>initial_array_capacity)) {
		half_size();
	}
}

/*
* Description: Empties the deque by resetting the member variables. The array is resized to the initial
*              capacity.
*/
template <typename Type>
void Resizable_deque<Type>::clear() {
	
	//Initializing all elements of current array to zero...

	for (int i = ifront; i < array_capacity; i++) {
		array[i] = 0;
	}
	for (int j = 0; j < ifront; j++) {
		array[j] = 0;
	}

	//Creating new temporary array of size 'intial_array_capacity'

	int tmp_size = initial_array_capacity;
	Type *tmp_array = new Type[tmp_size];

	//Copying all elements (of numeric value '0') of array to tmp_array

	for (int i = 0; i < tmp_size; i++) {
		tmp_array[i] = array[i];
	}

	//Re-assigning tmp_array to array and other member variables...

	ifront = initial_array_capacity;
	iback = -1;

	delete[] array;
	array = tmp_array;

	deque_size = 0;
	array_capacity = tmp_size;

}

/*********************************************************************
* *                       SUB SECTION - 4                         * *
* *            CUSTOM HELPER FUNCTIONS IMPLEMENTATION             * *
*********************************************************************/

/*
* Description: Returns true if the deque is full, false otherwise.
*/
template <typename Type>
bool Resizable_deque<Type>::full() const {
	
	return (deque_size == array_capacity);
}

/*
* Description: Returns the index of the element in front of the array
*/
template <typename Type>
int Resizable_deque<Type>::front_position() {

	return ifront;
}

/*
* Description: Returns the index of the element behind the array
*/
template <typename Type>
int Resizable_deque<Type>::back_position() {

	return iback;
}

/*
* Description: Retreives the value in the index of the array, passed as an argument
*/
template <typename Type>
Type Resizable_deque<Type>::retreive(int i) const {

	return array[i];
}

/*
* Description: Doubles the capacity of the array, ensuring order of all elements
*/
template <typename Type>
void Resizable_deque<Type>::double_size() {
	
	//Temporary array declaration and size initialization

	int tmp_size = 2*array_capacity;
	Type *tmp_array = new Type[tmp_size];

	//Initializing all elements of tmp_array as '0'

	for (int m = 0; m < tmp_size; m++) {
		tmp_array[m] = 0;
	}

	int ctr = 0;

	/* This while statement maps every element in array and copies it to tmp_array. The complexity
     * ensures that all elements are in order. */

	while (ctr != deque_size) {

		tmp_array[ctr] = array[ifront];
		ifront++;

		//If ifront is out of array bounds, loop back!

		if (ifront > (array_capacity - 1)) {
			ifront = 0;
		}

		ctr++;
	}

	//Re-initializing array to tmp_array and also respective member variables
	
	ifront = 0;
	iback = ctr - 1;

	delete[] array;
	array = tmp_array;

	deque_size = ctr;
	array_capacity = tmp_size;
}

/*
* Description: Halves the capacity of the array, ensuring order of all elements
*/
template <typename Type>
void Resizable_deque<Type>::half_size() {

	//Temporary array declaration and size initialization

	int tmp_size = array_capacity / 2;
	Type *tmp_array = new Type[tmp_size];

	//Initializing all elements of tmp_array as '0'

	for (int m = 0; m < tmp_size; m++) {
		tmp_array[m] = 0;
	}

	int ctr = 0;

	/* This while statement maps every element in array and copies it to tmp_array. The complexity
	* ensures that all elements are in order. */

	while (ctr != deque_size) {

		//Since, we are halving, we want to ensure all non-zero elements are copied only. Else there might be loss of data

		if (array[ifront] == 0) {

			ctr += 0;
		}
		else{

			tmp_array[ctr] = array[ifront];
			ctr++;
		}

		ifront++;

		//If ifront is out of array bounds, loop back!

		if (ifront >(array_capacity - 1)) {

			ifront = 0;
		}
		
	}
	
	//Re-initializing array to tmp_array and also respective member variables

	ifront = 0;
	iback = ctr - 1;

	delete[] array;
	array = tmp_array;

	deque_size = ctr;
	array_capacity = tmp_size;

}

/*********************************************************************
* *                       SUB SECTION - 5                         * *
* *                   FRIEND IMPLEMENTATION                       * *
*********************************************************************/

// You can modify this function however you want:  it will not be tested

template <typename T>
std::ostream &operator<<( std::ostream &out, Resizable_deque<T> const &list ) {
	out << "not yet implemented";

	return out;
}

#endif
