﻿/*****************************************
* uWaterloo User ID:  csubrama @uwaterloo.ca
* Submitted for ECE 250
* Department of Electrical and Computer Engineering
* University of Waterloo
* Calender Term of Submission: Winter 2018
*
* By submitting this file, I affirm that
* I am the author of all modifications to
* the provided code.
*
* The following is a list of uWaterloo User IDs of those students
* I had discussions with in preparing this project:
*    -Professor Harder: I looked at Prof. Harder's slides for reference
*    in making an adjacency matrix and it's various functionalities!
*
* The following is a list of uWaterloo User IDs of those students
* who helped me with this project (describe their help; e.g., debugging):
*    -
*****************************************/

#ifndef WEIGHTED_GRAPH_H
#define WEIGHTED_GRAPH_H

#ifndef nullptr
#define nullptr 0
#endif

#include <iostream>
#include <limits>
#include "Exception.h"

/*
* Weighted_graph
*
* Modified by: Chandrakiran Subramaniam
* Date of completion: 2018-04-02,
*
* Below is a summary of all functions associated with this class.
* You will notice that it has been documented in the format of an
* UML class diagram. Comments have been included before every func.
* implementation for the ease understanding :)
*
+ degree( in n:Integer ):Integer
+ edge_count():Integer
+ adjacent( in m:Integer, in n:Integer ):Real
+ distance( in m:Integer, in n:Integer ):Real
+ insert( in m:Integer, in n:Integer, in w:Real )
*/

class Weighted_graph {
private:
	
	//Member variables relating the properties of the graph
	//These include: Degree, length of edge (weight) and size of graph

	int *degree_vertex;
	int edgeWeight = 0;

	int size = 50;

	double **adjacencyMatrix;

	static const double INF;

	/*
	* vertex_table
	*
	* Made by: Chandrakiran Subramaniam
	* Date of completion: 2018-01-14, 00:51:46 EST
	*
	* This class is used to form a vertex table for the ease of implementation
	* of our algorithm and will be used by our queue. The properties of this 
	* class include
	* 
	* bool visited: Returns true if a vertex has been visited.
	* double distance: returns the shortest distance from original path
	* vertex_table *previous: points to the previous vertex
	*/

	class vertex_table {
	public:
		bool visited = false;
		double distance = INF;
		vertex_table *previous = nullptr;
	};

public:
	Weighted_graph(int = 50);
	~Weighted_graph();

	//Accessors to various graph properties...

	int degree(int) const;
	int edge_count() const;
	double adjacent(int, int) const;
	double distance(int, int);

	void insert(int, int, double);

	// Friends

	friend std::ostream &operator<<(std::ostream &, Weighted_graph const &);
};

/**************************************************************************
* * *                        SUB SECTION - 1                         * * *
* * *           CONSTRUCTORS & DESTRUCTORS IMPLEMENTATION            * * *
***************************************************************************/

/*Description: Construct an undirected graph with n vertices (by default, 50).           
*/
Weighted_graph::Weighted_graph(int n) {

	//If n ≤ 0, use n = 1.
	if (n <= 0) {
		n = 1;
	}

	size = n;

	//Declaring our Adjacency Matrix (Lower-triangular) to form our graph!
	//Allocating memory by adding appropriate amounts to the pointers

	adjacencyMatrix = new double *[n];
	adjacencyMatrix[0] = nullptr;
	adjacencyMatrix[1] = new double[n * (n - 1) / 2];

	for (int i = 2; i < n; i++) {
		adjacencyMatrix[i] = adjacencyMatrix[i - 1] + i - 1;
	}

	//Setting the weight of all edges to infinity (meaning they are all un-connected)

	for (int i = 0; i < n; i++) {
		for (int j = 0; j < i; j++) {
			adjacencyMatrix[i][j] = INF;
		}
	}

	//Allocating memory and initializing degree array to zero!

	degree_vertex = new int[n];
	for (int i = 0; i < n; i++) {
		degree_vertex[i] = 0;
	}

}

/*Description: Clean up any allocated memory.
*/
Weighted_graph::~Weighted_graph() {
	delete [] adjacencyMatrix;
	delete [] degree_vertex;
}


/**************************************************************************
* * *                        SUB SECTION - 2                         * * *
* * *                   ACCESSORS IMPLEMENTATION                     * * *
***************************************************************************/

/*Description: Returns the degree of the vertex n. Throw an illegal argument 
               exception if the argument does not correspond to an existing vertex.
*/
int Weighted_graph::degree(int d) const {
	if ((d < size) && (d>=0)) {
		return degree_vertex[d];
	}	
	throw illegal_argument();
}

/*Description: Returns the number of edges in the graph.
*/
int Weighted_graph::edge_count() const {
	return edgeWeight;
}

/*Description: Returns the weight of the edge connecting vertices m and n.
*/
double Weighted_graph::adjacent(int m, int n) const {

	//If the vertices are the same, return 0
	if (m == n) {
		return 0;
	}

	//If the m and n values are in valid range
	if ((m < size) && (m >= 0) && ((n < size) && (n >= 0))) {
		if (n < m) {
			return adjacencyMatrix[m][n];
		}
		else if (m < n) {
			return adjacencyMatrix[n][m];
		}
		else {
			return 0;
		}
	}
	else {

		//Throw an illegal argument exception if the arguments do not correspond to existing vertices
		throw illegal_argument();
	}

}

/*********************************************************************
* *                       SUB SECTION - 3                         * *
* *                  MUTATORS IMPLEMENTATION                      * *
*********************************************************************/

/*Description: Insert a vertex into the graph.
*/
void Weighted_graph::insert(int i, int j, double w) {
	if (w > 0 && i < size && j < size && i != j && i>=0 && j>=0) {
		//If the weight w > 0, add an edge between vertices i and j
		if (j < i) {
			if (adjacencyMatrix[i][j] == INF) {
				//In this case we are making a new edge

				degree_vertex[i] = degree_vertex[i] + 1;
				degree_vertex[j] = degree_vertex[j] + 1;
				edgeWeight++;
			}
			adjacencyMatrix[i][j] = w;
		}
		else
		{
			if (adjacencyMatrix[j][i] == INF) {
				//In this case we are making a new edge

				degree_vertex[i] = degree_vertex[i] + 1;
				degree_vertex[j] = degree_vertex[j] + 1;
				edgeWeight++;
			}
			adjacencyMatrix[j][i] = w;
		}
	}
	else {
		throw illegal_argument();
	}
}

/*Description: Return the shortest distance between vertices m and n. 
*/
double Weighted_graph::distance(int m, int n) {

	//If the m and n values are in valid range
	if ((m < size) && (m >= 0) && ((n < size) && (n >= 0))) {

		//We make a priority queue
		vertex_table *queue = new vertex_table[size];

		//Set distance of starting vertex to zero
		queue[m].distance = 0;

		for (int j = 0; j < size; j++) {

			//Traverse thorugh the graph to find the vertex with smallest weight

			int minI = -1;
			double minDst = INF;

			for (int i = 0; i < size; i++) {
				if (queue[i].visited == false && queue[i].distance < minDst) {
					minDst = queue[i].distance;
					minI = i;
				}
			}
			if (minI == -1) {
				for (int i = 0; i < size; i++) {
					if (queue[i].visited == false) {
						minI = i;
						break;
					}
				}
			}

			queue[minI].visited = true;

			//Update distance for all non-visited vertices!
			for (int i = 0; i < size; i++) {
				if (adjacent(minI, i) != INF && queue[i].visited == false && minI != i) {
					if (queue[i].distance > minDst + adjacent(minI, i)) {
						queue[i].distance = minDst + adjacent(minI, i);
					}
				}
			}
		}

		//After looping 'V' times, all vertices are visited!
		return queue[n].distance;

		delete[] queue;
	}
	else {
		//Throw an illegal argument exception if the arguments do not correspond to existing vertices
		throw illegal_argument();
	}

}

const double Weighted_graph::INF = std::numeric_limits<double>::infinity();

std::ostream &operator<<(std::ostream &out, Weighted_graph const &graph) {
	return out;
}

// Is an error showing up in ece250.h or elsewhere?
// Did you forget a closing '}' ?

#endif
